#std
import logging
from time import time
#gevent
from gevent.event import Event
#shaveet
from shaveet.utils import guid,LRU
from shaveet.consts import SYSTEM_ID,ADMIN_CHANNEL
from shaveet.config import MAX_MESSAGES_PER_CHANNEL,MAX_IDLE_CLIENT,COMET_TIMEOUT,MIN_ALIVE_TIME,MAX_CLIENTS_RECONNECT
import shaveet#for api

logger = logging.getLogger("shaveet.gc")

class Client(object):
  """
  Represents a client in shaveet,a client is identified by string called id and is unique per client.
  """
  __slots__ = ('id','channels','ts','create_ts','channels_event','is_waiting','key')

  def __init__(self,id):
    self.id = id
    self.channels = {}
    #timestamp of last action,used for GC
    self.ts = int(time())
    #when this object was created
    self.create_ts = int(time())
    self.channels_event = Event()
    self.is_waiting = False
    self.key = guid()

  def touch(self):
    self.ts = int(time())
    
  def get_subscribed_channels(self):
    return self.channels.keys()
    
  def add_channel(self,channel_name,cursor = False):
    existed = channel_exist(channel_name)
    chn = get_channel(channel_name) 
    if not channel_name in self.channels:
      if cursor and existed:#in case of reconnection set to old position
        self.channels[channel_name] = cursor
      else:#set to chn.id so we don't get old messages
        self.channels[channel_name] = chn.id - 1
    chn.add_client(self.id)
    #notifies that this client has a new channel
    self.channels_event.set()
    self.channels_event.clear()

  def remove_channel(self,channel_name,gc=False):
    if channel_name in self.channels:
      del self.channels[channel_name]
    get_channel(channel_name).remove_client(self.id,gc)
    
  def update_cursors(self,updates):
    for channel_name,messages in updates.iteritems():
      if channel_name in self.channels:
        self.channels[channel_name] = messages[-1].id
    
  def mark_waiting(self):
    self.is_waiting = True
    
  def clear_waiting(self):  
    self.is_waiting = False
    
  def is_active(self,now=False):
    #if now is provided use that to save a call to time
    if not now:
      now = int(time())
    idle_time = now - self.ts 
    alive_time = now - self.create_ts
    #if the client is idle for short period or waiting it is consider active
    return alive_time <= MIN_ALIVE_TIME or idle_time <= MAX_IDLE_CLIENT or (self.is_waiting and idle_time <= (10 + COMET_TIMEOUT))
    
  def remove_from_channels(self):
    for channel_name in self.channels.keys():
      self.remove_channel(channel_name,True)

#channel name -> Channel
_channels = {}
#client id -> Client
_clients = {}
#client id -> Client
_dead_clients = LRU(MAX_CLIENTS_RECONNECT)

###############
#  client     #
###############

def create_client(client_id):
  global _clients
  if not client_id in _clients:
    _clients[client_id] = Client(client_id)
  return _clients[client_id].key
  
def get_client(client_id):
  global _clients
  return _clients[client_id]  
  
def client_exists(client_id):
  global _clients
  return client_id in _clients
  
def all_clients():
  return _clients.copy()
  
def discard_client(client,kill=False):
  if client_exists(client.id):
    old_channels = client.channels.copy()
    client.remove_from_channels()
    del _clients[client.id]
    if kill:
      del client
    else:
      _dead_clients[client.id] = (client,old_channels)
    
    
def get_client_with_key(client_id_key):
  client_id,key = client_id_key.split("_!!_")
  if client_id in _dead_clients:
    #resurrect the client :)
    client,old_channels =  _dead_clients[client_id]
    _clients[client_id] = client
    for channel_name,cursor in old_channels.iteritems():
      client.add_channel(channel_name,cursor)
    del _dead_clients[client_id]
    logger.info("%s came back to life!",client_id)
  client = _clients[client_id]
  if client.key != key:
    raise KeyError("Wrong key for client")
  return client
###############
#  channel    #
###############
def get_channel(channel_name):
  from shaveet.channel import Channel
  global _channels
  if not channel_name in _channels:
    _channels[channel_name] = Channel(channel_name,MAX_MESSAGES_PER_CHANNEL) 
    shaveet.api.new_message(SYSTEM_ID,{"type":"new_channel","name":channel_name},ADMIN_CHANNEL,True)
  return _channels[channel_name]
  
def channel_exist(channel_name):
  global _channels
  return channel_name in _channels
    



from flask import Flask,request
from wsgi_jsonrpc.json_tools import ServerProxy
from uuid import uuid4
from json import dumps

app = Flask(__name__)

@app.route("/create_client")
def create_client():
  shaveet = ServerProxy("http://localhost:8082")
  client_id = request.args.get('name') + "_" + str(uuid4()).replace("-","") + 'S3CR37';
  key = shaveet.create_client(client_id)['result']
  shaveet.subscribe(client_id,'chat-room')
  return dumps({'key':key,'client_id':client_id})

@app.route("/new_message")
def new_message():
  shaveet = ServerProxy("http://localhost:8082")
  shaveet.new_message(request.args.get('client_id'),request.args.get('msg'),'chat-room',False);
  return "1"

if __name__ == "__main__":
  app.debug = True
  app.run()


from gevent.hub import getcurrent,get_hub
from uuid import uuid4

def guid():
  return str(uuid4()).replace("-","")

def maxfd():
    import resource
    return resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    
def setprocname(name):
  try:
    from setproctitle import setproctitle
    setproctitle(name)
  except:
    pass
  
def waitany(events):
  """
  wait until any of the provided events get triggered
  """
  switch = getcurrent().switch
  try:
    for ev in events:
      ev.rawlink(switch)
    return get_hub().switch()
  finally:
    for ev in events:
      ev.unlink(switch)
      
      
class IPMiddleware(object):
  def __init__(self,app,ips):
    self.app = app
    self.ips = set(ips)
    
  def __call__(self,env,start_response):
    
    if env.get('REMOTE_ADDR') in self.ips or not self.ips:
      return self.app(env,start_response)
    else:
      start_response('404 Not Found',[('Content-Type', 'text/html')])
      return ['Nothing here buddy']


class Node(object):
  __slots__ = ['prev', 'next', 'me']
  def __init__(self, prev, me):
    self.prev = prev
    self.me = me
    self.next = None

class LRU:
  """
  Implementation of a length-limited O(1) LRU queue.
  Built for and used by PyPE:
  http://pype.sourceforge.net
  Copyright 2003 Josiah Carlson.
  """
  def __init__(self, count, pairs=[]):
    self.count = max(count, 1)
    self.d = {}
    self.first = None
    self.last = None
    for key, value in pairs:
      self[key] = value
  def __contains__(self, obj):
    return obj in self.d
  def __getitem__(self, obj):
    a = self.d[obj].me
    self[a[0]] = a[1]
    return a[1]
  def __setitem__(self, obj, val):
    if obj in self.d:
      del self[obj]
    nobj = Node(self.last, (obj, val))
    if self.first is None:
      self.first = nobj
    if self.last:
      self.last.next = nobj
    self.last = nobj
    self.d[obj] = nobj
    if len(self.d) > self.count:
      if self.first == self.last:
        self.first = None
        self.last = None
        return
      a = self.first
      a.next.prev = None
      self.first = a.next
      a.next = None
      del self.d[a.me[0]]
      del a
  def __delitem__(self, obj):
    nobj = self.d[obj]
    if nobj.prev:
      nobj.prev.next = nobj.next
    else:
      self.first = nobj.next
    if nobj.next:
      nobj.next.prev = nobj.prev
    else:
      self.last = nobj.prev
    del self.d[obj]
  def __iter__(self):
    cur = self.first
    while cur != None:
      cur2 = cur.next
      yield cur.me[1]
      cur = cur2
  def iteritems(self):
    cur = self.first
    while cur != None:
      cur2 = cur.next
      yield cur.me
      cur = cur2
  def iterkeys(self):
    return iter(self.d)
  def itervalues(self):
    for i,j in self.iteritems():
      yield j
  def keys(self):
    return self.d.keys()


#std
import logging
#3rd
from gevent import Greenlet,sleep
#shaveet
from shaveet.config import MAX_CLIENTS_GC,CLIENT_GC_INTERVAL
from shaveet.lookup import all_clients,discard_client

logger = logging.getLogger("shaveet.gc")

class ClientGC(Greenlet):
  """
  this greenthread collects the clients that are no longer active
  """
  def run(self):
    while True:
      logger.info("ClientGC:processing clients")
      client_processed = 0
      for client_id,client in all_clients().iteritems():
        if not client.is_active():
          logger.debug("ClientGC:collecting id:%s,ts:%d,waiting:%s",client.id,client.ts,client.is_waiting)
          discard_client(client)
        #process in chuncks of MAX_CLIENTS_GC,sleep(0) means yeild to next greenlet
        client_processed+=1
        if client_processed % MAX_CLIENTS_GC == 0:
            sleep(0)
      logger.info("ClientGC:sleeping")
      sleep(CLIENT_GC_INTERVAL)


#std
from __future__ import with_statement
from os import sep,path
from optparse import OptionParser
import sys,logging,logging.handlers,traceback
#3rd party
from static import Cling
from daemon import DaemonContext
from gevent import wsgi
#fix path
sys.path.append(sep.join(path.dirname(path.abspath(__file__)).split(sep)[:-1]))
#shaveet
from shaveet import config,api,gc
from shaveet.utils import maxfd,setprocname

BASE_PATH = path.dirname(path.abspath(__file__))
logger_wsgi = logging.getLogger("wsgi_jsonrpc")
logger = logging.getLogger("shaveet.gc")
static_app = Cling(path.dirname(path.abspath(__file__)))

def wsgi_main(env, start_response):
  try:
    if env['PATH_INFO'] == '/':
      return api.handle(env,start_response)
    elif env['PATH_INFO'].startswith('/message_updates'):
      return [api.message_updates(env,start_response)]
    elif env['PATH_INFO'].startswith('/static'):
      return static_app(env, start_response)
    else:
      start_response('404 Not Found', [('Content-Type', 'text/plain')])
      return 'Not Found\r\n'
  except GeneratorExit:#Python 2.5 fix
    pass
  except Exception,e:
    start_response('500 Internal server error', [('Content-Type', 'text/plain')])
    logger.error(traceback.format_exc())
    return 'Internal server error\r\n'

def main(options):
  server = wsgi.WSGIServer(('',options.port or config.PORT), wsgi_main)
  client_gc = gc.ClientGC()
  try:
    client_gc.start()
    server.serve_forever()
  except KeyboardInterrupt:
    server.stop()

def cli_start():
  setprocname("shaveet")
  parser = OptionParser()
  parser.add_option("-d", "--daemon", dest="daemon",help="run as daemon",action="store_true",default=False)
  parser.add_option("-p", "--port",dest="port",help="which port to listen",type="int",default=0)
  (options, args) = parser.parse_args()
  
  formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
  #setup handlers
  fh = logging.handlers.RotatingFileHandler(config.LOG_PATH,maxBytes = 1024 * 1024,backupCount=3)
  ch = logging.StreamHandler()
  #set formmaters
  fh.setFormatter(formatter)
  ch.setFormatter(formatter)
  logger_wsgi.setLevel(logging.INFO)
  logger.setLevel(logging.INFO)

  if options.daemon:
    with DaemonContext(files_preserve=range(maxfd() + 2)):      
      logger.addHandler(fh)
      logger_wsgi.addHandler(fh)
      logger.info("starting daemon..")
      main(options)
  else:
    logger.setLevel(logging.DEBUG)
    logger_wsgi.setLevel(logging.DEBUG)
    logger.addHandler(ch)
    logger_wsgi.addHandler(ch)
    main(options)
    
if __name__ == "__main__":
    cli_start()
    



# -*- coding: utf-8 -*-
#std
from datetime import datetime
from urllib import unquote
from urlparse import parse_qs
try:
  from json import loads,dumps
except ImportError:#Python 2.5
  from simplejson import loads,dumps
#3rd
import wsgi_jsonrpc
#binfire
from shaveet.utils import IPMiddleware
from shaveet.config import COMET_TIMEOUT,IPS
from shaveet.lookup import get_client,get_client_with_key,get_channel,channel_exist,discard_client,create_client
from shaveet.channel import get_updates

def message_updates(env,start_response):
  """
  handles clients long polling,wait for updates and return them or timeout
  """
  #parse request,looks like ?client_id=....&client_id=....&callback=...
  qs = parse_qs(env['QUERY_STRING'])
  #client_id is client_id;key
  client_ids = [id.decode('utf8') for id in qs["client_id"]]
  callback = qs["callback"][0]
  try:
    cursors = []
    clients = [get_client_with_key(client_id_key) for client_id_key in client_ids]
    #get the updates,blocks if no updates found
    updates = get_updates(clients)
    #if updates found update the clients cursors
    for client in clients:
      client.update_cursors(updates)
    start_response('200 OK', [('Content-Type', 'application/x-javascript'),('Access-Control-Allow-Origin','*')])
    #generates the response like: callback({"channel_1":[[message],[message]],"channel_2":[[message]]})
    return "".join([callback,"(",dumps(dict((channel_name,[[message.id,message.message,message.client_id] for message in messages]) 
                                                                                  for channel_name,messages in updates.iteritems())),");"])
  except KeyError:#one of the clients doesn't exists
    start_response('404 Not Found', [('Content-Type', 'text/plain')])
    return "Not Found\r\n"  


def subscribe(client_id,channel_name):
  """
  subscribe the client to the channel
  
  client_id - id of the client
  channel_name - name of the channel
  """
  get_client(client_id).add_channel(channel_name)
  return 1

def unsubscribe(client_id,channel_name):
  """
  unsubscribe the client to the channel
  
  client_id - id of the client
  channel_name - name of the channel
  """
  get_client(client_id).remove_channel(channel_name)
  return 1
  
def subscribe_many(channels):
  """
  subscribe many clients to many channels
  
  channels - a dict like this {
    channel_name:[client_id,client_id2]
  }
  """
  for channel,clients in channels.iteritems():
    for client_id in clients:
      subscribe(client_id,channel)

def unsubscribe_many(channels):
  """
  unsubscribe many clients to many channels
  
  channels - a dict like this {
    channel_name:[client_id,client_id2]
  }
  """
  for channel,clients in channels.iteritems():
    for client_id in clients:
      unsubscribe(client_id,channel)
      
def unsubscribe_all_channel(channel_name,exclude):
  """
  unsubscribe all the clients from the channel except those in the exclude list
  
  channel_name - name of the channel
  exclude - a list of clients to exclude from this operation
  """
  unsubscribe_many({channel_name:filter(lambda x: not x in exclude,get_channel_clients(channel_name))})

def new_message(client_id,message,channel_name,system):
  """
  publish a new message into the channel message queue
  
  client_id - the publishing client id
  message - a string or object to publish,if it is a object it is sent as a json string
  channel_name - the name of the channel to publish this message
  system - true for system messages(like unsubscribe because of GC)
  """
  if not (type(message) is str or type(message) is unicode):
    message = dumps(message)
  get_channel(channel_name).new_message(client_id,message,system)
  return 1
  
def get_channel_clients(channel_name):
  """
  get all the clients subscribed to the channel
  
  channel_name - name of the channel
  """
  
  return get_channel(channel_name).get_active_clients()
  
def get_client_channels(client_id):
  """
  get all the clients subscribed to the channel
  
  channel_name - name of the channel
  """
  
  return get_client(client_id).get_subscribed_channels()
  
def kill_client(client_id):
  return discard_client(get_client(client_id),True)
  
#this is the wsgi application entry point
api_wsgi = wsgi_jsonrpc.WSGIJSONRPCApplication(methods=[subscribe,subscribe_many,unsubscribe,unsubscribe_many,unsubscribe_all_channel,new_message,
                                                        get_channel_clients,get_client_channels,kill_client,create_client])
                                                        
handle = IPMiddleware(api_wsgi,IPS)




from setuptools import setup

setup(name='shaveet', 
      version='0.1.5', 
      packages=['shaveet'],
      package_data={'shaveet': ['static/*.js']},
      install_requires = ['gevent>=0.13','python-daemon','wsgi-jsonrpc','static'],
      extras_require = {'process title':  ["setproctitle"]},
      author = "Uriel Katz",
      author_email = "uriel.katz@gmail.com",
      description = "Shaveet comet server",
      license = "MIT",
      keywords = "comet gevent server",
      url = "https://github.com/urielka/shaveet",
      download_url = "https://github.com/urielka/shaveet/downloads",
      zip_safe = False,
      entry_points = '''
        [console_scripts]
        shaveet = shaveet.main:cli_start
        ''',
)



from time import time
from json import dumps
#gevent
from gevent.event import Event
from gevent import Timeout
#binfire
from shaveet.lookup import get_client,get_channel,channel_exist,client_exists
from shaveet.config import COMET_TIMEOUT
from shaveet.utils import waitany

MESSAGE_INDEX = 0
CLIENT_ID_INDEX = 1
ID_INDEX = 2

class ChannelCursor(object):
  """
  represent the position of a client in the channel(i.e. which messages he got up to this point
  """
  
  __slots__ = ('channel','cursor')
  
  def __init__(self,channel,cursor):
    self.channel = channel
    self.cursor = cursor
    
class Message(object):
  """
  represent a message in the channel.
  """
  
  __slots__ = ('client_id','message','id','system')
  
  def __init__(self,client_id,message,id,system):
    self.client_id = client_id
    self.message = message
    self.id = id
    self.system = system

class Channel(object):
  """
  Represent a channel in shaveet,a channel is something that clients can subscribe to and publish messages to.
  Each subscribed client get all the messages published to the channel.
  Each channel is identified by a name unique to that channel
  """
  
  __slots__ = ('clients','messages','new_message_event','max_messages','id','name')
  
  def __init__(self,name,max_messages):
    self.name = name
    #a cache of recent messages limited by max_messages
    self.messages = []
    #a event that fire on every new message
    self.new_message_event = Event()
    self.max_messages = max_messages
    self.id = 1
    self.clients = set()
    
  def add_client(self,client_id):
    self.clients.add(client_id)
    self.new_message(client_id,dumps({"type":"subscribe"}),True)
    
  def remove_client(self,client_id,gc=False):
    self.clients.discard(client_id)
    self.new_message(client_id,dumps({"type":"unsubscribe","gc":gc}),True)

  def new_message(self,client_id,message,system=False):
    self.messages.append(Message(client_id,message,self.id,system))
    self.id += 1
    #trim list,remove old messages
    if len(self.messages) > self.max_messages:
      self.messages = self.messages[-self.max_messages:]
    self.new_message_event.set()
    self.new_message_event.clear()

  def get_updates(self,cursor):
    """
    get updates for the channel,if there is no updates returns a event for notification
    """
    if not self.messages or cursor >= self.messages[-1].id:
      return self.new_message_event
    
    return [message for message in self.messages if message.id > cursor]
    
  def get_active_clients(self):
    now = int(time())#saves call to time
    return [client_id for client_id in self.clients
              if client_id and not "admin" in client_id and client_exists(client_id) and get_client(client_id).is_active(now)]
              
  def __hash__(self):
    return hash(self.name)
    

def get_updates(clients):
  """
  get all updates from the channels and waits until one channel have new information
  """
  #channel name -> messages
  updates = {}
  #events to wait for
  events = []
  #cursors for each channel
  cursors = []
  for client in clients:
    client.touch()
    for channel_name,cursor in client.channels.iteritems():
      if channel_exist(channel_name):
        cursors.append(ChannelCursor(get_channel(channel_name),cursor))
  #check for available messages
  for cc in cursors:
    update = cc.channel.get_updates(cc.cursor)
    if type(update) is Event:
      events.append(update)
    else:
      updates[cc.channel.name] = update
  #found messages,so just return them
  if len(updates) > 0:
    return updates
  #no messages,so wait for any event
  elif len(events) > 0:
    fired = False
    try:
      #mark the client as waiting so the GC doesn't collect them
      for client in clients:
        client.mark_waiting()
      #wait for any event to get fired up to COMET_TIMEOUT seconds
      with Timeout(COMET_TIMEOUT,False) as timeout:
        waitany(events + [client.channels_event for client in clients])
        fired = True
    finally:
      #clear the marking
      for client in clients:
        client.touch()
        client.clear_waiting()
    if fired:
      return get_updates(clients)
    else:#timeout
      return {}
  #no channels
  else:
    return {}


import os.path

PORT = 8082
MAX_MESSAGES_PER_CHANNEL = 10
COMET_TIMEOUT = 25#Seconds
MAX_IDLE_CLIENT = 10#Seconds
MIN_ALIVE_TIME = 60
MAX_CLIENTS_GC = 1000
MAX_CLIENTS_RECONNECT = 100000#Amount of clients to keep for reconnection
MAX_CLIENT_RECONNECT_IDLE = 300#Seconds
CLIENT_GC_INTERVAL = 10

IPS = ['127.0.0.1']
LOG_PATH = os.path.expanduser('~/shaveet.log')